import os

import time

import numpy as np
from transformers import Wav2Vec2Processor, Wav2Vec2ForCTC

from cltl.asr.api import ASR
from cltl.asr.util import store_wav


class Wav2Vec2ASR(ASR):
    def __init__(self, model_id: str, sampling_rate: int, storage: str = None):
        self.processor = Wav2Vec2Processor.from_pretrained(model_id)
        self.model = Wav2Vec2ForCTC.from_pretrained(model_id)

        self.sampling_rate = sampling_rate

        self._storage = storage

    def speech_to_text(self, audio: np.array, sampling_rate: int) -> str:
        if not sampling_rate:
            sampling_rate = self.sampling_rate

        raw_audio = self._resample(audio, sampling_rate)
        if self._storage:
            store_wav(raw_audio, sampling_rate, str(os.path.join(self._storage, f"asr-{time.time()}.wav")))

        representation = self.processor(raw_audio, sampling_rate=self.sampling_rate, return_tensors="pt").input_values
        token_logits = self.model(representation).logits
        predicted_tokens = np.argmax(token_logits.detach().numpy(), axis=2)

        return self.processor.decode(predicted_tokens[0])

    def _resample(self, audio, sampling_rate):
        if not audio.dtype == np.int16:
            raise ValueError(f"Invalid sample depth {audio.dtype}, expected np.int16")

        if sampling_rate != self.sampling_rate:
            raise ValueError(f"Unsupported sampling rate: {sampling_rate}, expected {self.sampling_rate}")

        if audio.ndim == 1 or audio.shape[1] == 1:
            return audio.squeeze()

        if audio.ndim > 2 or audio.shape[1] > 2:
            raise ValueError(f"audio must have at most two channels, shape was {audio.shape}")

        return audio.mean(axis=1, dtype=np.int16).ravel()



import os
import shutil
import tempfile
import time

import numpy as np
from speechbrain.pretrained import EncoderDecoderASR

from cltl.asr.api import ASR
from cltl.asr.util import store_wav


class SpeechbrainASR(ASR):
    def __init__(self, model_id: str, storage: str = None, model_dir: str = None):
        self.processor = EncoderDecoderASR.from_hparams(source=model_id, savedir=model_dir)
        self._storage = storage if storage else tempfile.mkdtemp()

    def clean(self):
        shutil.rmtree(self._storage)

    def speech_to_text(self, audio: np.array, sampling_rate: int) -> str:
        wav_file = str(os.path.join(self._storage, f"asr-{time.time()}.wav"))
        try:
            store_wav(audio, sampling_rate, wav_file)

            return self.processor.transcribe_file(wav_file)
        finally:
            os.remove(wav_file)



from dataclasses import dataclass

from cltl_service.backend.schema import TextSignalEvent
from emissor.representation.container import Index
from emissor.representation.scenario import Modality


@dataclass
class AsrTextSignalEvent(TextSignalEvent):
    confidence: float
    audio_segment: Index

    @classmethod
    def create(cls, signal_id: str, timestamp: float, transcript: str, confidence: float, audio_segment: Index):
        return cls(cls.__name__, signal_id, timestamp, Modality.TEXT, [], transcript, confidence, audio_segment)


import logging
import uuid
from typing import Callable

import numpy as np
import time
from cltl.backend.api.storage import STORAGE_SCHEME
from cltl.backend.source.client_source import ClientAudioSource
from cltl.backend.spi.audio import AudioSource
from cltl.combot.infra.config import ConfigurationManager
from cltl.combot.infra.event import Event, EventBus
from cltl.combot.infra.resource import ResourceManager
from cltl.combot.infra.topic_worker import TopicWorker
from cltl_service.vad.schema import VadMentionEvent
from emissor.representation.container import Index
from emissor.representation.scenario import Modality

from cltl.asr.api import ASR
from cltl_service.asr.schema import AsrTextSignalEvent

logger = logging.getLogger(__name__)


CONTENT_TYPE_SEPARATOR = ';'


class AsrService:
    @classmethod
    def from_config(cls, asr: ASR, event_bus: EventBus, resource_manager: ResourceManager,
                    config_manager: ConfigurationManager):
        config = config_manager.get_config("cltl.asr")

        def audio_loader(url, offset, length) -> AudioSource:
            return ClientAudioSource.from_config(config_manager, url, offset, length)

        return cls(config.get("vad_topic"), config.get("asr_topic"), asr, audio_loader, event_bus, resource_manager)

    def __init__(self, vad_topic: str, asr_topic: str, asr: ASR, audio_loader: Callable[[str, int, int], AudioSource],
                 event_bus: EventBus, resource_manager: ResourceManager):
        self._asr = asr
        self._audio_loader = audio_loader
        self._event_bus = event_bus
        self._resource_manager = resource_manager
        self._vad_topic = vad_topic
        self._asr_topic = asr_topic

        self._topic_worker = None

    @property
    def app(self):
        return None

    def start(self, timeout=30):
        self._topic_worker = TopicWorker([self._vad_topic], self._event_bus, provides=[self._asr_topic],
                                         resource_manager=self._resource_manager, processor=self._process)
        self._topic_worker.start().wait()

    def stop(self):
        if not self._topic_worker:
            pass

        self._topic_worker.stop()
        self._topic_worker.await_stop()
        self._topic_worker = None

    def _process(self, event: Event[VadMentionEvent]):
        payload = event.payload
        segment: Index = payload.mention.segment[0]
        url = f"{STORAGE_SCHEME}:{Modality.AUDIO.name.lower()}/{segment.container_id}"

        with self._audio_loader(url, segment.start, segment.stop - segment.start) as source:
            transcript = self._asr.speech_to_text(np.concatenate(tuple(source.audio)), source.rate)

        asr_event = self._create_payload(transcript, payload)
        self._event_bus.publish(self._asr_topic, Event.for_payload(asr_event))

    def _create_payload(self, transcript, payload):
        signal_id = str(uuid.uuid4())

        return AsrTextSignalEvent.create(signal_id, time.time(), transcript, 1.0, payload.mention.segment)



import numpy as np
import sounddevice as sd
import soundfile


def store_wav(frames, sampling_rate, save=None):
    if not isinstance(frames, np.ndarray):
        audio = np.concatenate(frames)
    else:
        audio = frames

    if save:
        soundfile.write(save, audio, sampling_rate)
    else:
        sd.play(audio, sampling_rate)
        sd.wait()



import abc

import numpy as np


class ASR(abc.ABC):
    def speech_to_text(self, audio: np.ndarray, sampling_rate: int) -> str:
        """
        Transcribe the provided audio sample to text.

        Parameters
        ----------
        audio : np.array
            The binary audio data containing speech.
            The provided np.array must be either of shape (n), (n, 1) for mono
            or (n, 2) for stereo input, where n is the number of samples
            contained in the audio.
        sampling_rate : int
            The sampling rate of the audio data. If supported by the
            implementation, a falsy value may be passed to use a default sample
            rate. It is not recommended for implementations to provide the
            sampling rate as default argument, as this can lead to hard to
            detect errors if sampling rates don't match.

        Returns
        -------
        str
            Text transcript of the audio input.
        """
        raise NotImplementedError()


from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION", "r") as fh:
    version = fh.read().strip()

setup(
    name='cltl.asr',
    version=version,
    package_dir={'': 'src'},
    packages=find_namespace_packages(include=['cltl.*', 'cltl_service.*'], where='src'),
    data_files=[('VERSION', ['VERSION'])],
    url="https://github.com/leolani/cltl-asr",
    license='MIT License',
    author='CLTL',
    author_email='t.baier@vu.nl',
    description='ASR for Leolani',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.7',
    install_requires=['numpy'],
    extras_require={
        "impl": [
            "cffi",
            "importlib_resources",
            "jiwer",
            "sounddevice",
            "soundfile",
            "torch",
            "transformers",
            "speechbrain"
        ],
        "service": [
            "cltl.backend",
            "cltl.combot",
            "cltl.vad",
            "emissor",
            "requests",
        ]
    },
)

